import os
from pathlib import Path

import time
import random
import numpy as np

import parse_intent
import utils
import dataset_io as dio
import parameter_io as pio
import population_funcs


class Universe(object):
    """
    Create the universe of the simulation.
    """

    def __init__(self,
                 dataset_fn=None,
                 world_param_fn=None,
                 species_param_fns=None,
                 world_param_dict={},
                 species_param_dicts=[{}],
                 custom_module_fns=None,
                 current_time=0,
                 end_time=10,
                 dataset_dir='datasets/',
                 pad_zeros=0,
                 file_extension='.txt',
                 seed=None):
        """
        Initialize universe based on either parameter files or saved datasets.

        Parameters
        ----------
        dataset_fn : str
            Filename of saved organism + world dataset.
        world_param_fn : str
            Filename of world parameter file.
        species_param_fns : list of str
            List of filenames of species parameter files.
        world_param_dict : dict
            Dictionary containing initial world parameters.
        species_param_dicts : list of dict
            List of dictionaries containing initial species parameters.
        custom_module_fns : list of str
            List of filenames of external python scripts containing custom
            behaviors.
        current_time : int
            Current time of simulation.
        end_time : int
            End time of simulation.
        dataset_dir : str
            Directory path for saving all world and organism datasets.
        pad_zeros : int
            Number of zeroes to pad in dataset filenames.
        file_extension : str
            File extension for saving dataset files. Should generally be '.txt'
            or '.json'.
        seed : int, optional
            Random seed for the simulation.
        """
        # Set random seeds for the entire simulation
        if seed is not None:
            random.seed(seed)
            np.random.seed(seed)

        self.start_timestamp = time.time()
        self.last_timestamp = self.start_timestamp

        self.dataset_fn = dataset_fn
        self.world_param_fn = world_param_fn
        self.species_param_fns = species_param_fns
        self.custom_module_fns = custom_module_fns

        self.world_param_dict = world_param_dict
        self.species_param_dicts = species_param_dicts

        if self.custom_module_fns is not None:
            self.custom_module_fns = [os.path.abspath(path)
                                      for path in self.custom_module_fns
                                      if os.path.isfile(path)]

        self.dataset_dir = Path(dataset_dir)
        self.dataset_dir.mkdir(exist_ok=True)

        self.current_time = current_time
        self.end_time = end_time
        self.pad_zeros = pad_zeros
        # while (self.end_time - self.current_time) >= 10 ** self.pad_zeros:
        #     self.pad_zeros += 1
        self.file_extension = file_extension

        self.population_dict, self.world = self.initialize()
        self.species_names = sorted(list(self.population_dict.keys()))
        self.intent_list = []

    def initialize(self):
        """
        Initialize world and organisms in the universe, from either saved
        datasets or from parameter files (and subsequently writing the
        initial time step to file).

        Returns
        -------
        population_dict : dict
            Dict of organisms at the beginning of the simulation.
        world : World
            World at the beginning of the simulation.
        """
        if self.dataset_fn is not None:
            # Set up entire universe based on saved dataset
            population_dict, world = dio.load_universe(self.dataset_fn)

        if self.world_param_fn is not None:
            # Set up entire world based on parameter file
            world = pio.load_world(fn=self.world_param_fn)
        else:
            world = pio.load_world(init_dict=self.world_param_dict)

        if self.species_param_fns is not None:
            # Set up all organisms based on species specifications
            population_dict = pio.load_species(
                                  fns=self.species_param_fns,
                                  init_world=world,
                                  custom_module_fns=self.custom_module_fns)
        else:
            population_dict = pio.load_species(
                                  init_dicts=self.species_param_dicts,
                                  init_world=world,
                                  custom_module_fns=self.custom_module_fns)

        output_fn = (
            self.dataset_dir / 'ds{}'.format(str(self.current_time)
                                             .zfill(self.pad_zeros))
        ).with_suffix(self.file_extension)
        dio.save_universe(population_dict, world, output_fn)

        return population_dict, world

    def step(self):
        """
        Steps through one time step, iterating over all organisms and
        computing new organism states. Saves all organisms and the world
        to file at the end of each step.
        """
        # Increment time step
        self.current_time += 1

        # This is just updating the age, not evaluating whether an organism
        # is at death, since organism actions should be evaluated based on
        # the current state. Age needs to be updated so that every organism
        # in intent list has the correct age.
        organism_list = population_funcs.get_organism_list(self.population_dict)
        t_organism_list = [organism.clone_self()._update_age()
                           for organism in organism_list
                           if organism.alive]
        position_hash_table = (population_funcs
                               .hash_by_position(t_organism_list))

        # intent_list is a list of lists, one list per organism in the current
        # time step
        self.intent_list = []
        for organism in organism_list:
            if organism.alive:
                # currently t_organism_list isn't used by any actions...
                self.intent_list.append(
                    organism.step(population_funcs.get_population_dict(t_organism_list,
                                                                       self.species_names),
                                  self.world,
                                  position_hash_table=position_hash_table)
                )

        # Parse intent list and ensure it is valid
        self.population_dict = parse_intent.parse(self.intent_list,
                                                  self.population_dict)

        # Potential changes to the world would go here
        self.world.step()

        output_fn = (
            self.dataset_dir / 'ds{}'.format(str(self.current_time)
                                             .zfill(self.pad_zeros))
        ).with_suffix(self.file_extension)
        dio.save_universe(self.population_dict, self.world, output_fn)

    def current_info(self, verbosity=1, expanded=True):
        total_num = sum([self.population_dict[species]['statistics']['total']
                         for species in self.species_names])

        pstring = 't = %s' % (self.current_time)
        if verbosity >= 1:
            if expanded:
                pstring = (
                    '... t = %s\n'
                    % str(self.current_time).zfill(self.pad_zeros)

                    + '    Number of organisms: %s\n'
                      % total_num
                )
            else:
                rt_pstring = 't = %s: %s organisms' % (self.current_time,
                                                       total_num)
        if verbosity >= 4:
            if expanded:
                for species_name in self.species_names:
                    pstring += (
                        '    %s: %d organisms\n'
                        % (species_name, self.population_dict[species_name]['statistics']['total'])
                    )
            else:
                rt_pstring = rt_pstring + ' ('
                for i, species_name in enumerate(self.species_names):
                    rt_pstring += str(self.population_dict[species_name]['statistics']['total'])
                    if i != len(self.species_names) - 1:
                        rt_pstring += ':'
                rt_pstring += ')'

        if verbosity >= 2:
            now = time.time()
            last_time_diff = now - self.last_timestamp
            self.last_timestamp = now
            if expanded:
                pstring += (
                    '    Time elapsed since last time step: %s\n'
                    % utils.time_to_string(last_time_diff)
                )
            else:
                pstring = rt_pstring + (
                    ' (%s)'
                    % (utils.time_to_string(last_time_diff))
                )
        if verbosity >= 3:
            start_time_diff = now - self.start_timestamp
            if expanded:
                pstring += (
                    '    Time elapsed since start: %s\n'
                    % utils.time_to_string(start_time_diff)
                )
            else:
                pstring = rt_pstring + (
                    ' (%s; %s)'
                    % (utils.time_to_string(last_time_diff),
                       utils.time_to_string(start_time_diff))
                )

        return pstring

    def run(self, verbosity=1, expanded=True):
        print(self.current_info(verbosity=verbosity, expanded=expanded))
        while self.current_time < self.end_time:
            self.step()
            print(self.current_info(verbosity=verbosity, expanded=expanded))


# At its simplest, the entire executable could just be written like this
if __name__ == '__main__':
    universe = Universe()
    universe.run()



import uuid
import copy
import imp
import sys
import random

import default_fields
from utils import cast_to_list
from organism_behavior import movement, reproduction, drinking, eating, action


class Organism(object):
    """
    A basic organism structure for all species.
    """

    def __init__(self, init_dict={}):
        """
        Create a new organism from a dictary of parameters. The dictionary
        is specified in blossom.default_fields.
        """
        # Set up defaults based on organism parameters
        for (field, default) in default_fields.organism_fields.items():
            setattr(self, field, init_dict.get(field, default))

        # Set up custom fields provided in initialization dictionary
        init_keys = set(init_dict.keys())
        default_keys = set(default_fields.organism_fields.keys())
        for custom_field in (init_keys - default_keys):
            setattr(self, custom_field, init_dict[custom_field])

        # Set unique id for organism
        if self.organism_id is None:
            # self.organism_id = str(uuid.uuid4())
            self.organism_id = self.get_new_id()

        # Set current water level for uninitialized organism
        if self.drinking_type is not None and self.water_current is None:
            self.water_current = self.water_initial

        # Set current food level for uninitialized organism
        if self.eating_type is not None and self.food_current is None:
            self.food_current = self.food_initial

        # Import custom modules / paths
        if self.custom_module_fns is not None:
            self._custom_modules = []
            for i, path in enumerate(cast_to_list(self.custom_module_fns)):
                temp_module = imp.load_source('%s' % i, path)
                self._custom_modules.append(temp_module)

    def to_dict(self):
        """
        Convert Organism to dict.
        """
        organism_vars = vars(self)
        public_vars = {key: val
                       for key, val in organism_vars.items()
                       if not key.startswith('_')}
        return public_vars

    def get_new_id(self):
        """
        Generates pseudo-random ID for the organism, seeded by the universe.
        """
        return str(uuid.UUID(int=random.getrandbits(128)))

    @classmethod
    def clone(cls, organism):
        """
        Makes a new Organism object identical to the current one.

        Parameters
        ----------
        organism : Organism
            Organism to copy.

        Returns
        -------
        new_organism : Organism
            Copied organism.
        """
        new_organism = cls(organism.to_dict())
        # Use copy module to properly handle mutable lists
        new_organism.ancestry = copy.copy(new_organism.ancestry)
        new_organism.position = copy.copy(new_organism.position)
        return new_organism

    def clone_self(self):
        """
        Clone this organism.
        """
        return self.clone(self)

    def get_child(self, other_parent=None):
        """
        Creates an Organism object with similar properties to self, and can
        add another parent if it exists. Note that this doesn't assume
        anything about how much food / water the child is left with, so these
        should be set with custom / default reproduction methods.

        Parameters
        ----------
        other_parent : Organism
            Parent that reproduces with self to produce the child.

        Returns
        -------
        child : Organism
            Generated child.
        """
        child = self.clone_self()
        child.age = 0
        child.organism_id = child.get_new_id()
        if other_parent is None:
            child.ancestry.append(self.organism_id)
        else:
            child.ancestry.append([self.organism_id, other_parent.organism_id])
        child.last_action = None
        return child

    def update_parameter(self, parameter, value, method='set', in_place=False, original=None):
        """
        Update a specific parameter of the organism.

        Parameters
        ----------
        parameter : string
            Parameter to update.
        value
            Value with which to update.
        method : string
            Method types are: 'set', 'add', 'subtract', 'append'.
        in_place : bool
            If True, modifies self, otherwise, copy organism and return new
            Organism object.
        original : Organism or None
            Original organism we are changing. If it is the original,
            clone organism so that we aren't editing the original.

        Returns
        -------
        updated_organism : Organism
            Organism object with updated parameter.
        """
        if self is original or not in_place:
            updated_organism = self.clone_self()
        else:
            updated_organism = self
        attribute = getattr(updated_organism, parameter)
        if method == 'set':
            attribute = value
        elif method == 'add':
            attribute += value
        elif method == 'subtract':
            attribute -= value
        elif method == 'append':
            attribute.append(value)
        else:
            sys.exit('Invalid update method!')
        setattr(updated_organism, parameter, attribute)
        return updated_organism

    def move(self, population_dict, world, position_hash_table=None):
        """
        Method for handling movement. Searches through custom methods and
        built-in movement methods.

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table : dict
            Dictionary mapping positions to available organisms.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's movement.
        """
        try:
            if self.movement_type is None:
                sys.exit('No movement type defined!')
            elif self.custom_module_fns is not None:
                for custom_module in self._custom_modules:
                    if hasattr(custom_module, self.movement_type):
                        return getattr(custom_module, self.movement_type)(
                            self,
                            population_dict,
                            world,
                            position_hash_table=position_hash_table
                        )
            return getattr(movement, self.movement_type)(
                self,
                population_dict,
                world,
                position_hash_table=position_hash_table
            )
        except AttributeError as e:
            raise AttributeError(
                str(e)
                + '. Check that \'%s\' method is indeed ' % self.movement_type
                + 'contained within the provided custom modules and that the '
                + 'method is written correctly.'
            ).with_traceback(sys.exc_info()[2])

    def reproduce(self, population_dict, world, position_hash_table=None):
        """
        Method for handling reproduction. Searches through custom methods
        and built-in reproduction methods.

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table : dict
            Dictionary mapping positions to available organisms.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's
            reproduction. For example, this would include both parent and
            child organisms.

        """
        try:
            if self.reproduction_type is None:
                sys.exit('No reproduction type defined!')
            elif self.custom_module_fns is not None:
                for custom_module in self._custom_modules:
                    if hasattr(custom_module, self.reproduction_type):
                        return getattr(custom_module, self.reproduction_type)(
                            self,
                            population_dict,
                            world,
                            position_hash_table=position_hash_table
                        )
            return getattr(reproduction, self.reproduction_type)(
                self,
                population_dict,
                world,
                position_hash_table=position_hash_table
            )
        except AttributeError as e:
            raise AttributeError(
                str(e) + '. Check '
                + 'that \'%s\' method is indeed ' % self.reproduction_type
                + 'contained within the provided custom modules and that the '
                + 'method is written correctly.'
            ).with_traceback(sys.exc_info()[2])

    def drink(self, population_dict, world, position_hash_table=None):
        """
        Method for handling drinking. Searches through custom methods and
        built-in drinking methods.

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table : dict
            Dictionary mapping positions to available organisms.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's drinking.

        """
        try:
            if self.drinking_type is None:
                sys.exit('No drinking type defined!')
            elif self.custom_module_fns is not None:
                for custom_module in self._custom_modules:
                    if hasattr(custom_module, self.drinking_type):
                        return getattr(custom_module, self.drinking_type)(
                            self,
                            population_dict,
                            world,
                            position_hash_table=position_hash_table
                        )
            return getattr(drinking, self.drinking_type)(
                self,
                population_dict,
                world,
                position_hash_table=position_hash_table
            )
        except AttributeError as e:
            raise AttributeError(
                str(e)
                + '. Check that \'%s\' method is indeed ' % self.drinking_type
                + 'contained within the provided custom modules and that the '
                + 'method is written correctly.'
            ).with_traceback(sys.exc_info()[2])

    def eat(self, population_dict, world, position_hash_table=None):
        """
        Method for handling eating. Searches through custom methods and
        built-in eating methods.

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table : dict
            Dictionary mapping positions to available organisms.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's eating.

        """
        try:
            if self.eating_type is None:
                sys.exit('No eating type defined!')
            elif self.custom_module_fns is not None:
                for custom_module in self._custom_modules:
                    if hasattr(custom_module, self.eating_type):
                        return getattr(custom_module, self.eating_type)(
                            self,
                            population_dict,
                            world,
                            position_hash_table=position_hash_table
                        )
            return getattr(eating, self.eating_type)(
                self,
                population_dict,
                world,
                position_hash_table=position_hash_table
            )
        except AttributeError as e:
            raise AttributeError(
                str(e)
                + '. Check that \'%s\' method is indeed ' % self.eating_type
                + 'contained within the provided custom modules and that the '
                + 'method is written correctly.'
            ).with_traceback(sys.exc_info()[2])

    def act(self, population_dict, world, position_hash_table=None):
        """
        Method that decides and calls an action for the current timestep.
        Searches through custom methods and built-in movement methods.
        The action method specifically selects an action to take, from "move",
        "reproduce", "drink", and "eat". Then the appropriate instance method
        from this class is executed to yield the final list of affect
        organisms.

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table : dict
            Dictionary mapping positions to available organisms.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's action.

        """
        action_name = None
        try:
            if self.custom_module_fns is not None:
                for custom_module in self._custom_modules:
                    if hasattr(custom_module, self.action_type):
                        action_name = getattr(custom_module, self.action_type)(
                            self,
                            population_dict,
                            world,
                            position_hash_table=position_hash_table
                        )
            if action_name is None:
                action_name = getattr(action, self.action_type)(
                    self,
                    population_dict,
                    world,
                    position_hash_table=position_hash_table
                )
        except AttributeError as e:
            raise AttributeError(
                str(e)
                + '. Check that \'%s\' method is indeed ' % self.action_type
                + 'contained within the provided custom modules and that the '
                + 'method is written correctly.'
            ).with_traceback(sys.exc_info()[2])

        self.last_action = action_name

        affected_organisms = cast_to_list(getattr(self, action_name)(
            population_dict,
            world,
            position_hash_table=position_hash_table
        ))

        # Ensure this organism is included in affected_organisms
        already_included = False
        for org in affected_organisms:
            if org.organism_id == self.organism_id:
                already_included = True
        if not already_included:
            # e.g. unknown reason for exlusion, so default to death
            self.die('unknown', in_place=True)
            affected_organisms.append(self)

        return self, affected_organisms

    def _update_age(self):
        """
        Increments age by 1.
        """
        self.age += 1
        return self

    def _update_water(self):
        """
        Updates health parameters relevant to water consumption.

        Decreases current water level based on metabolism, and increments
        time without water accordingly. Note that organisms die of thirst if
        this reaches the maximum time without water.
        """
        self.water_current -= self.water_metabolism

        if self.water_current > self.water_capacity:
            self.water_current = self.water_capacity
        elif self.water_current <= 0:
            self.water_current = 0
            self.time_without_water += 1
        elif self.time_without_water > 0:
            self.time_without_water = 0
        return self

    def _update_food(self):
        """
        Updates health parameters relevant to food consumption.

        Decreases current food level based on metabolism, and increments
        time without food accordingly. Note that organisms die of hunger if
        this reaches the maximum time without food.
        """
        self.food_current -= self.food_metabolism

        if self.food_current > self.food_capacity:
            self.food_current = self.food_capacity
        elif self.food_current <= 0:
            self.food_current = 0
            self.time_without_food += 1
        elif self.time_without_food > 0:
            self.time_without_food = 0
        return self

    def is_at_death(self, cause):
        """
        Check various conditions for death.

        Parameters
        ----------
        cause : str
            Potential cause of this organism's death.

        Returns
        -------
        is_dead : bool
            Returns True if organism is dead from the specified cause, False
            otherwise.
        """
        if cause == 'old_age':
            return self.age > self.max_age
        elif cause == 'thirst':
            return (self.drinking_type is not None
                    and self.time_without_water > self.max_time_without_water)
        elif cause == 'hunger':
            return (self.eating_type is not None
                    and self.time_without_food > self.max_time_without_food)
        else:
            sys.exit('Invalid cause!')

    def die(self, cause, in_place=False, original=None):
        """
        Method that "kills" organism.

        Parameters
        ----------
        cause : str
            Cause of this organism's death.
        in_place : bool
            If True, modifies self, otherwise, copy organism and return new
            Organism object.

        Returns
        -------
        dead_organism : Organism
            New "dead" state of this organism.
        """
        updated_organism = self.update_parameter('alive',
                                                 False,
                                                 in_place=in_place,
                                                 original=original)
        updated_organism = updated_organism.update_parameter('age_at_death',
                                                             self.age,
                                                             in_place=in_place,
                                                             original=original)
        updated_organism = updated_organism.update_parameter('cause_of_death',
                                                             cause,
                                                             in_place=in_place,
                                                             original=original)
        return updated_organism

    def step(self, population_dict, world, position_hash_table=None, do_action=True):
        """
        Steps through one time step for this organism. Reflects changes
        based on actions / behaviors and updates to health parameters.

        Returns a list of organisms that the action produced (either new or
        altered organisms).

        Parameters
        ----------
        population_dict : dict of Organisms
            Dict of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.
        position_hash_table: dict
            Dictionary of organisms "hashed" by position.
        do_action: bool
            If True, this organism will act, otherwise, it will not.

        Returns
        -------
        affected_organisms : list of Organisms
            List of organisms affected by this organism's actions
            or health. This could be an updated version of this organism,
            especially if the organism dies during the time step, but could
            also be multiple other organisms affected by actions (i.e. children
            from reproduction).

        """
        # Create new Organism object / reference and update age
        organism = self.clone_self()._update_age()

        if organism.is_at_death('old_age'):
            # This organism is at death from old age
            # Design choice to prevent action if organism is at old age limit
            return [organism.die('old_age')]

        if do_action:
            organism, affected_organisms = organism.act(
                population_dict,
                world,
                position_hash_table=position_hash_table
            )
        else:
            affected_organisms = [organism]

        for organism in affected_organisms:
            # Checks whether alive after action
            if organism.alive:
                if organism.is_at_death('old_age'):
                    organism.die('old_age', in_place=True)

                # Update and check water / food status
                if organism.drinking_type is not None:
                    organism._update_water()
                    if organism.is_at_death('thirst'):
                        organism.die('thirst', in_place=True)

                if organism.eating_type is not None:
                    organism._update_food()
                    if organism.is_at_death('hunger'):
                        organism.die('hunger', in_place=True)

        return affected_organisms

    def step_without_acting(self):
        """
        Steps through one time step without acting for this organism.

        Returns
        -------
        organism
            Note that this returns an Organism object, not a list.

        """
        return self.step(population_dict=None,
                         world=None,
                         position_hash_table=None,
                         do_action=False)[0]



import random
import population_funcs


def parse(intent_list, population_dict):
    """
    Determine whether the intent list is valid and fix it in the event of
    conflicts.

    Parameters
    ----------
    intent_list : list of lists of Organisms
        List of lists of organisms with proposed organism states,
        after each organism has 'acted'. Length equals number of organisms in
        the current time step.
    population_dict : dict of Organisms
        Dict of current organisms

    Returns
    -------
    updated_population_dict : dict of Organisms
        Dict of updated organisms, where organism states that conflict between
        intent_list and population_dict are resolved.

    Conflicts may be cases in which an organism has different states in the
    intent list, perhaps arrising from the actions of other organisms that
    somehow effect its state. This method resolves those conflicts, so that
    there is only one organism with a given organism id present in the final
    output list at all times.
    """
    # TODO: Figure out exactly how this should be controlled -- on the scale of
    # the universe, the world, or the organisms itself
    updated_list = []

    id_org_dict = {}
    for species in population_dict:
        for org in population_dict[species]['organisms']:
            id_org_dict[org.organism_id] = org

    new_organism_ids = set()

    # Randomly sample organism steps to select. Only use sets for conditionals,
    # add to saved structures using lists and dicts (since key order is
    # preserved)
    for organism_set in random.sample(intent_list, len(intent_list)):
        set_ids = set(org.organism_id for org in organism_set)
        if len(new_organism_ids & set_ids) == 0:
            updated_list.extend(organism_set)
            new_organism_ids.update(set_ids)

    # Add back organisms whose steps were not chosen (and increment status)
    for id in id_org_dict.keys():
        if id not in new_organism_ids:
            org = id_org_dict[id]
            if org.alive:
                updated_list.append(org.step_without_acting())

    return population_funcs.get_population_dict(updated_list,
                                                population_dict.keys())



import default_fields


class World(object):
    """
    World class for the environment of the simulation.
    """

    def __init__(self, init_dict={}):
        """
        Create a new world from a dictary of parameters. The dictionary
        is specified in blossom.default_fields.
        """
        # Set up defaults based on world parameters
        for (field, default) in default_fields.world_fields.items():
            setattr(self, field, init_dict.get(field, default))

        # Set up custom fields provided in initializaiton dictionary
        init_keys = set(init_dict.keys())
        default_keys = set(default_fields.world_fields.keys())
        for custom_field in (init_keys - default_keys):
            setattr(self, custom_field, init_dict[custom_field])

    def to_dict(self):
        """
        Convert World to dict.
        """
        world_vars = vars(self)
        public_vars = {key: val
                       for key, val in world_vars.items()
                       if not key.startswith('_')}
        return public_vars

    def step(self):
        self.current_time += 1



"""
Load information from parameter files and construct world and
organism objects at the initial timestep.
"""

import glob
import configparser
import json
import random
import copy

from utils import cast_to_list
from world import World
from organism import Organism
import default_fields


def load_world(fn=None, init_dict={}):
    """
    Load world from either parameter file or dictionary and construct
    initial World object.

    Parameters
    ----------
    fn : str
        Input filename of parameter file.
    init_dict : dict
        Dictionary containing world parameters.

    Returns
    -------
    world : World
        World object constructed from the parameter file.
    """
    if fn:
        return load_world_from_param_file(fn)
    else:
        return load_world_from_dict(init_dict)


def load_world_from_dict(init_dict):
    world_init_dict = {}
    for (field, default) in default_fields.world_fields.items():
        world_init_dict[field] = init_dict.get(field, default)
    # Check world size parameters, set dimensionality value accordingly
    if type(world_init_dict['world_size']) is int:
        world_init_dict['world_size'] = [world_init_dict['world_size']]
    else:
        assert type(world_init_dict['world_size']) is list
        assert len(world_init_dict['world_size']) <= 2
    world_init_dict['dimensionality'] = len(world_init_dict['world_size'])
    for i in world_init_dict['world_size']:
        assert type(i) is int

    # Check that field list sizes are consistent with each other
    field_lengths = []
    fields = []
    for field in ['water', 'food', 'obstacles']:
        if world_init_dict[field]:
            assert type(world_init_dict[field]) is list
            # Check each element of these fields is either int or float (can
            # include math.inf, which is a float)
            for i in world_init_dict[field]:
                if type(i) is list:
                    for j in i:
                        assert type(j) in [int, float]
                else:
                    assert type(i) in [int, float]
            field_lengths.append(len(world_init_dict[field]))
            fields.append(field)
    if len(field_lengths) >= 2:
        assert field_lengths[0] == field_lengths[1]
    if len(field_lengths) == 3:
        assert field_lengths[1] == field_lengths[2]

    # Check that field list sizes are consistent with the world size
    if len(fields) >= 1:
        field = fields[0]
        assert world_init_dict['world_size'][0] == len(world_init_dict[field])
        if len(field) == 2:
            assert world_init_dict['world_size'][1] \
                == len(world_init_dict[field][0])

    return World(world_init_dict)


def load_world_from_param_file(fn):
    """
    Load world parameter file and construct initial World object.

    Parameters
    ----------
    fn : str
        Input filename of parameter file.

    Returns
    -------
    world : World
        World object constructed from the parameter file.
    """

    env_file = glob.glob(fn)
    if len(env_file) == 0:
        raise IndexError('There is no environment configuration file in the '
                         + 'current directory.')
    if len(env_file) > 1:
        print('There are multiple environment configuration files in the '
              + 'current directory. There should only be one environment '
              + 'configuration file. Selecting the first provided..')
    env_file = env_file[0]

    world_init_dict = {}
    # Load from config file
    config_world = configparser.ConfigParser()
    config_world.read(env_file)

    # environment_filename: str, or 'None'
    environment_filename = config_world.get('Overall Parameters',
                                            'environment_filename')
    if environment_filename != 'None':
        with open(environment_filename, 'r') as f:
            initial_environment_dict = json.load(f)

        world_init_dict['water'] = initial_environment_dict['water']
        world_init_dict['food'] = initial_environment_dict['food']
        world_init_dict['obstacles'] = initial_environment_dict['obstacles']

        if type(world_init_dict['water'][0]) is list:
            world_init_dict['dimensionality'] = 2
            world_init_dict['world_size'] = [len(world_init_dict['water']),
                                             len(world_init_dict['water'][0])]
        else:
            world_init_dict['dimensionality'] = 1
            world_init_dict['world_size'] = [len(world_init_dict['water'])]

    else:
        # world_size: space delimited ints, or 'None'
        # example: world_size = 10 10
        world_size = config_world.get('Overall Parameters', 'world_size')
        if world_size == 'None':
            # this is probably harder to do,
            # since we'd have to allow for infinite bounds...
            world_size = None
            return -1
        else:
            world_size = [int(L) for L in world_size.split()]
            world_init_dict['world_size'] = world_size
            world_init_dict['dimensionality'] = len(world_size)

            if len(world_size) == 2:
                blank_vals = [[0 for x in range(world_size[1])]
                              for x in range(world_size[0])]
            elif len(world_size) == 1:
                blank_vals = [0 for x in range(world_size)[0]]
            else:
                raise ValueError

            world_init_dict['water'] = copy.deepcopy(blank_vals)
            world_init_dict['food'] = copy.deepcopy(blank_vals)
            world_init_dict['obstacles'] = copy.deepcopy(blank_vals)

    return World(world_init_dict)


def load_species(fns=None,
                 init_dicts=[{}],
                 init_world=World({}),
                 custom_module_fns=[]):
    """
    Load organisms from available species parameter files or dictionaries.

    Parameters
    ----------
    fns : list of str
        Input filenames of species parameter files. Different species get
        different species parameter files, from which the individual organisms
        are initialized.
    init_dicts : list of dict
        Parameter dicts for each species.
    init_world : World
        Initial World instance for this Universe.
    custom_module_fns : list of str
        List of external Python scripts containing custom organism
        behaviors. :mod:`blossom` will search for methods within each
        filename included here.

    Returns
    -------
    population_dict : dict of Organisms
        A dict of Organism objects constructed from the parameter file.

    """
    if fns:
        return load_species_from_param_files(fns,
                                             init_world,
                                             custom_module_fns)
    else:
        return load_species_from_dict(init_dicts,
                                      init_world,
                                      custom_module_fns)


def create_organisms(species_init_dict,
                     init_world=World({}),
                     position_callback=None):
    '''
    Make organism list from an species_init_dict either provided directly or
    scraped from parameter file. All organisms are from a single species.
    '''
    organism_list = []
    list_field_keys = []
    for key in species_init_dict.keys():
        if type(species_init_dict[key]) is list:
            if key != 'custom_module_fns':
                list_field_keys.append(key)

    # Generate all organisms
    initial_population = species_init_dict['population_size']
    for i in range(initial_population):
        organism_init_dict = {
            key: val for key, val in species_init_dict.items()
            if not key == 'population_size'
        }
        for key in list_field_keys:
            organism_init_dict[key] = organism_init_dict[key][i]
        if 'initial_positions' in species_init_dict:
            position = organism_init_dict['initial_positions']
        elif position_callback:
            position = position_callback(init_world.world_size)
        elif 'position_callback' in species_init_dict:
            position = (species_init_dict['position_callback']
                        (init_world.world_size))
        else:
            # Vary organism location randomly
            position = []
            for i in range(init_world.dimensionality):
                position.append(random.randrange(0,
                                                 init_world.world_size[i]))
        organism_init_dict['position'] = position

        # Add organism to organism list
        organism_list.append(Organism(organism_init_dict))

    return organism_list


def load_species_from_dict(init_dicts,
                           init_world,
                           custom_module_fns=None):
    """
    Create a list of organisms loaded from Python dicts.

    Parameters
    ----------
    init_dicts : list of dict
        Input species dictionaries from which the individual organisms
        are initialized. Each dictionary is for a different species.
    init_world : World
        Initial World instance for this Universe.
    custom_module_fns : list of str
        List of external Python scripts containing custom organism
        behaviors. :mod:`blossom` will search for methods within each
        filename included here.

    Returns
    -------
    population_dict : dict of Organisms
        A dict of Organism objects constructed from the parameter file.

    """
    init_dicts = cast_to_list(init_dicts)

    population_dict = {}
    for init_dict in init_dicts:

        species_init_dict = {}

        # Set up defaults based on species parameters
        for (field, default) in default_fields.species_fields.items():
            species_init_dict[field] = init_dict.get(field, default)

        # Set up custom fields provided in initialization dictionary
        init_keys = set(init_dict.keys())
        default_keys = set(default_fields.species_fields.keys())
        for custom_field in (init_keys - default_keys):
            species_init_dict[custom_field] = init_dict[custom_field]

        assert 'population_size' in init_dict
        population_size = init_dict['population_size']
        assert type(population_size) is int
        species_init_dict['population_size'] = population_size

        assert type(species_init_dict['species_name']) == str

        for field in ['movement_type',
                      'reproduction_type',
                      'drinking_type',
                      'eating_type',
                      'action_type']:
            assert (species_init_dict[field] is None
                    or type(species_init_dict[field]) is str)

        for field in ['dna_length']:
            assert type(species_init_dict[field]) is int

        if species_init_dict['drinking_type']:
            assert (
                species_init_dict['water_capacity'] is not None
                and species_init_dict['water_initial'] is not None
                and species_init_dict['water_metabolism'] is not None
                and species_init_dict['water_intake'] is not None
                and species_init_dict['max_time_without_water'] is not None
            )
        if species_init_dict['eating_type']:
            assert (
                species_init_dict['food_capacity'] is not None
                and species_init_dict['food_initial'] is not None
                and species_init_dict['food_metabolism'] is not None
                and species_init_dict['food_intake'] is not None
                and species_init_dict['max_time_without_food'] is not None
            )

        for field in ['max_age',
                      'max_time_without_food',
                      'max_time_without_water',
                      'mutation_rate',
                      'food_capacity',
                      'food_initial',
                      'food_metabolism',
                      'food_intake',
                      'water_capacity',
                      'water_initial',
                      'water_metabolism',
                      'water_intake']:
            if type(species_init_dict[field]) is list:
                assert (len(species_init_dict[field]) == population_size)
            elif type(species_init_dict[field]) in [int, float]:
                species_init_dict[field] = ([species_init_dict[field]]
                                            * population_size)
            else:
                assert species_init_dict[field] is None

        if not species_init_dict['custom_module_fns']:
            # Track custom method file paths
            species_init_dict['custom_module_fns'] = custom_module_fns

        if 'initial_positions' in init_dict:
            assert len(init_dict['initial_positions']) == population_size
            for position in init_dict['initial_positions']:
                assert len(position) == init_world.dimensionality
            species_init_dict['initial_positions'] \
                = init_dict['initial_positions']

        species_organism_list = create_organisms(species_init_dict, init_world)

        # Populate population dict with relevant bulk stats and organism lists
        population_dict[species_init_dict['species_name']] = {
            'statistics': {
                'total': population_size,
                'alive': population_size,
                'dead': 0
            },
            'organisms': species_organism_list
        }

    return population_dict


def load_species_from_param_files(fns,
                                  init_world,
                                  custom_module_fns=None):
    """
    Load all available species parameter files.

    Parameters
    ----------
    fns : list of str
        Input filenames of species parameter files. Different species get
        different species parameter files, from which the individual organisms
        are initialized.
    init_world : World
        Initial World instance for this Universe.
    custom_module_fns : list of str
        List of external Python scripts containing custom organism
        behaviors. :mod:`blossom` will search for methods within each
        filename included here.

    Returns
    -------
    population_dict : dict of Organisms
        A dict of Organism objects constructed from the parameter file.

    """

    # Find organism filenames, can be a list of patterns
    fns = cast_to_list(fns)
    org_files = [fn for pattern in fns for fn in glob.glob(pattern)]

    # Initialize list of dictionaries to hold all organism parameters
    # Each dictionary contains parameters for a single species
    population_dict = {}
    for i, org_file in enumerate(org_files):
        # Initialize temporary dict to store species parameters
        species_init_dict = {}

        # Parameters that must be str or None
        param_str = ['species_name',
                     'movement_type',
                     'reproduction_type',
                     'drinking_type',
                     'action_type',
                     'eating_type']

        # Parameters that must be int
        param_int = ['population_size',
                     'dna_length']

        # Parameters that must be int or 'None' or 'inf'
        param_int_none_inf = ['max_age',
                              'max_time_without_food',
                              'max_time_without_water',
                              'mutation_rate',
                              'food_capacity',
                              'food_initial',
                              'food_metabolism',
                              'food_intake',
                              'water_capacity',
                              'water_initial',
                              'water_metabolism',
                              'water_intake']

        # Parameters that must be float
        param_float = ['proportion_m',
                       'proportion_f',
                       'proportion_a',
                       'proportion_h']

        # Parameters that must be bool
        param_bool = ['can_mutate']

        # Load from config file
        config_org = configparser.ConfigParser()
        config_org.read(org_file)

        # Cycle through all parameters in the config file,
        # converting them to proper types as specifed above
        for section in config_org.sections():
            for (key, val) in config_org.items(section):

                if key in param_str:
                    pass
                elif key in param_int:
                    val = int(val)
                elif key in param_int_none_inf:
                    if val != 'None' and type(val) not in [int, float]:
                        val = float(val)
                elif key in param_float:
                    val = float(val)
                elif key in param_bool:
                    val = (val == 'True')
                else:
                    print(key, val)
                    raise NameError('Key is not a valid parameter')
                # ensure 'None' parameters are set to None
                if val == 'None':
                    val = None
                species_init_dict[key] = val

        # Track custom method file paths
        species_init_dict['custom_module_fns'] = custom_module_fns

        species_organism_list = create_organisms(species_init_dict, init_world)

        # Populate population dict with relevant bulk stats and organism lists
        population_dict[species_init_dict['species_name']] = {
            'statistics': {
                'total': species_init_dict['population_size'],
                'alive': species_init_dict['population_size'],
                'dead': 0
            },
            'organisms': species_organism_list
        }

    return population_dict



def hash_by_id(organism_list):
    """
    Simple hashing by organism id over a list of organisms.
    """
    hash_table = {}
    for organism in organism_list:
        id = organism.organism_id
        if id in hash_table:
            hash_table[id].append(organism)
        else:
            hash_table[id] = [organism]
    return hash_table


def hash_by_position(organism_list):
    """
    Simple hashing by organism position over a list of organisms.
    """
    hash_table = {}
    for organism in organism_list:
        position = tuple(organism.position)
        if position in hash_table:
            hash_table[position].append(organism)
        else:
            hash_table[position] = [organism]
    return hash_table


def organism_filter(organism_list, *conditions):
    """
    Selects organisms from organism list according to a set of conditions.
    Each condition should be a function that receives an Organism object as
    input and returns a boolean as output.

    Example:
    organism_filter(
        population_dict['prey1']['organisms'],
        lambda organism: (organism.alive)
    """
    remaining_list = organism_list
    for condition in conditions:
        assert callable(condition)
        remaining_list = filter(condition, remaining_list)
    return list(remaining_list)


def organism_list_copy(organism_list):
    return [organism.clone_self() for organism in organism_list]


def get_organism_list(population_dict):
    """
    Constructs organism list from population dict data structure.
    """
    organism_list = []
    for species in population_dict:
        organism_list.extend(population_dict[species]['organisms'])
    return organism_list


def get_population_dict(organism_list, species_names):
    """
    Constructs population dict from organism list data structure.
    """
    population_dict = {
        species: {
            'statistics': {
                'total': 0,
                'alive': 0,
                'dead': 0
            },
            'organisms': []
        }
        for species in species_names
    }
    for organism in organism_list:
        population_dict[organism.species_name]['organisms'].append(organism)
        population_dict[organism.species_name]['statistics']['total'] += 1
        if organism.alive:
            population_dict[organism.species_name]['statistics']['alive'] += 1
        else:
            population_dict[organism.species_name]['statistics']['dead'] += 1
    return population_dict



"""
Load information from a certain dataset, e.g. to resume a simulation, and
write world and organism data back to file.
"""

import copy
import json

from world import World
from organism import Organism


def load_universe(fn):
    """
    Load dataset file from JSON.

    Parameters
    ----------
    fn : str
        Input filename of saved universe dataset.

    Returns
    -------
    population_dict : dict
        A dict of Organism objects reconstructed from the saved dataset.
    world : World
        World object reconstructed from the saved dataset.
    """
    with open(fn, 'r') as f:
        universe_dict = json.load(f)

    world = World(universe_dict['world'])

    population_dict_json = universe_dict['population']
    population_dict = {}
    for species in population_dict_json:
        population_dict[species] = {}
        population_dict[species]['statistics'] = copy.deepcopy(population_dict_json[species]['statistics'])
        population_dict[species]['organisms'] = [
            Organism(organism_dict)
            for organism_dict in population_dict_json[species]['organisms']
        ]

    return population_dict, world


def save_universe(population_dict, world, fn):
    """
    Save population_dict and world to file in JSON format.

    Parameters
    ----------
    population_dict : dict
        Dict of Organisms to write to file.
    world : World
        World attributes to write to file.
    fn : str
        Output filename of saved universe dataset.
    """
    population_dict_json = {}
    for species in population_dict:
        population_dict_json[species] = {}
        population_dict_json[species]['statistics'] = copy.deepcopy(population_dict[species]['statistics'])
        population_dict_json[species]['organisms'] = [
            organism.to_dict()
            for organism in population_dict[species]['organisms']
        ]

    universe_dict = {
        'population': population_dict_json,
        'world': world.to_dict()
    }

    with open(fn, 'w') as f:
        json.dump(universe_dict, f, indent=2)



import uuid
import math


def pure_replication(organism, population_dict, world, position_hash_table=None):
    """
    Replace organism with two organism with similar parameters.
    Essentially, only differences in parameters are organism id,
    ancestry, age, and water / food levels.
    """

    new_organism_list = []

    # Generate new organisms
    for i in range(2):
        child = organism.get_child()
        if organism.drinking_type is not None:
            child.update_parameter('water_current',
                                   math.floor(organism.water_current / 2),
                                   in_place=True)
        if organism.eating_type is not None:
            child.update_parameter('food_current',
                                   math.floor(organism.food_current / 2),
                                   in_place=True)
        new_organism_list.append(child)
    new_organism_list.append(organism.die('replication'))
    return new_organism_list



"""
Common utilities used throughout :mod:`blossom`
"""

import math


def cast_to_list(x):
    """
    Make a list out of the input if the input isn't a list.
    """
    if type(x) is list:
        return x
    else:
        return [x]


def time_to_string(seconds):
    """
    Convert time in seconds to the most reasonable representation.
    """
    if seconds < 1:
        return '%1.3f s' % seconds
    elif seconds < 60:
        return '%.2f s' % seconds
    elif seconds < 3600:
        minutes = math.floor(seconds / 60)
        seconds -= minutes * 60
        return '%d m %04.1f s' % (minutes, seconds)
    else:
        hours = math.floor(seconds / 3600)
        seconds -= hours * 3600
        minutes = math.floor(seconds / 60)
        seconds -= minutes * 60
        return '%d h %02d m %02d s' % (hours, minutes, seconds)



"""
blossom is a package for simulating evolution
"""

__version__ = "1.3.0"

import sys
import os
sys.path.append(os.path.dirname(__file__))

from universe import Universe
from organism import Organism
from world import World
from default_fields import world_fields, \
    specific_organism_fields, species_fields, \
    organism_fields
import dataset_io
import parameter_io

import organism_behavior
import world_generator
import utils
import population_funcs
